#include <Arduino.h>
#include <EEPROM.h>
#include "DATA.h"
#include "MENU.h"
#include "PINS.h"
#include "LCD.h"
#include "Button.h"
#include "RTC.h"
#include "GRAPH.h"
#include "IMU.h"
#include "CFG.h"

DATA data;

tRollingAverage::tRollingAverage()
{
  Reset();
}

void tRollingAverage::Reset()
{
  m_iDataIndex = 0;
  m_iSum = 0;
  m_pDataPoints[0] = 0;
  m_bFirst = true;
}

int tRollingAverage::DataPoint(int Data)
{
  // returns the abs value of Data's deviation from the running average
  // adds Data to the running average
  if (m_bFirst)
  {
    m_bFirst = false;
    // seed them all to the first data point
    for (int Idx = 0; Idx < kSize; Idx++)
    {
      m_pDataPoints[Idx] = Data;
    }
    m_iSum = Data*kSize;
    return 0;
  }
  else
  {
    m_iSum -= m_pDataPoints[m_iDataIndex];
    m_iSum += Data;
    m_pDataPoints[m_iDataIndex++] = Data;
    m_iDataIndex = m_iDataIndex % kSize;
    return abs((m_iSum / kSize) - Data);
  }
}

///////////////////////////////////////////////////////////////////////////////////////

void DATA::Init()
{
  m_iPageIndex = 0;
  memset(&m_CurrentData, 0, sizeof(m_CurrentData));
  if (btn2Red.IsDown())
  {
    // reset if red held at startup
    ResetEEPROM();
  }
}

bool DATA::CheckAutoStart()
{
  if (cfg.m_AutoOn && (rtc.ReadMinute() % 15) == 0)
  {
    bool bWeekDay = (rtc.ReadDayOfWeek() > 2);
    // only check on the quarters
    byte Hour;
    byte Quarter;
    cfg.UnPackTime(bWeekDay?cfg.m_WeekOn:cfg.m_WeekendOn, Hour, Quarter);
    rtc.ReadTime();
    // and only an exact match
    return (Hour == rtc.m_Hour24 && Quarter*15 == rtc.m_Minute);
  }
  return false;
}

// While running
// Green: Green = Advance
//        Red/Timeout = Resume
// Red:   Red = Stop
//        Green/Timeout = Resume
void DATA::Start()
{
  bool Logging = true;
  StartLogging();
  bool Match;
  int LatestPage = GetLatestDayPage();
  if (PageIsToday(LatestPage))
  {
    // re-use it
    m_iPageIndex = LatestPage;
  }
  else
  {
    // go to the next
    m_iPageIndex = LatestPage + 1;
    m_iPageIndex %= DATA_NUM_PAGES;
  }
  
  lcd.ShowGraphics(false);
  lcd.Clear();
  lcd.BigFont(0, 0, "Good Night! \x02");
  delay(2500);
  lcd.BacklightOn(false);
  lcd.Clear();
  do
  {
    unsigned long Now = millis();
    if (btn1Green.CheckButtonPress())
    {
      // green while logging, show graph
      GraphData();
      // wait for btn or 5s
      if (menu.WaitForButtonOrTimeout(5000) == Menu::eGreen)
      {
        // Green, shift fwd to now
        StartLogging();
      }
      // red, or timeout resume
      lcd.BacklightOn(false);
      lcd.Clear();
    }
    else if (btn2Red.CheckButtonPress())
    {
      // red while logging, show graph
      GraphData();
      // wait for btn or 5s
      if (menu.WaitForButtonOrTimeout(5000) == Menu::eRed)
      {
        // red, stop
        break;
      }
      else
      {
        // green, or timeout resume
        lcd.BacklightOn(false);
        lcd.Clear();
      }
    }
    
    if ((Now - m_iLastReadingMS) >= 250)
    {
      if (!TakeReading())
      {
        break;
      }
    }
  } while (true);
  StopLogging();
}

void DATA::StartLogging()
{
  rtc.ReadTime();
  m_CurrentData.m_iBaseReadingBy4 = cfg.m_iBaseReadingBy4;
  m_CurrentData.m_iThresholdPercent = cfg.m_iThresholdPercent;
  m_CurrentData.m_iStartDayOfWeek = rtc.m_DayOfWeek;
  m_CurrentData.m_iStartHour24 = rtc.m_Hour24;
  m_CurrentData.m_iStartMinute = rtc.m_Minute;
  m_CurrentData.m_iStartDate = rtc.m_DayOfMonth;
  m_CurrentData.m_iStartMonth = rtc.m_Month;
  m_CurrentData.m_iSampleCount = 0; 
  
  m_iLastReadingMS = millis();
  m_iLastReadingMinute = rtc.m_Minute;
  m_iReadingMinutes = 0;

  m_iSamplePercentSum = 0;
  m_iSamplePercentCount = 0;
  m_AvgX.Reset();
  m_AvgY.Reset();
  m_AvgZ.Reset();
  
  m_iMaxInactivity = 0;
  m_iInactivityWindowSum = 0;
  m_iInactivityWindowCount = 0;
}

void DATA::StopLogging()
{
  lcd.Clear();
  WriteData();
}

bool DATA::LoadLast()
{
  // load the page with the latest date
  int Page = GetLatestDayPage();
  if (Page != -1)
  {
    return ReadData(Page);
  }
  return false;
}

bool DATA::LoadPrev()
{
  // load the page prior to the current one
  int LastPage = GetLatestDayPage();
  int PrevPage = m_iPageIndex;
  if (PrevPage != -1)
  {
    PrevPage--;
    if (PrevPage < 0)
      PrevPage = DATA_NUM_PAGES - 1;
    if (PrevPage != LastPage)  // not back at start
    {
      return ReadData(PrevPage);
    }
  }
  return false;
}

void DATA::GraphData()
{
  lcd.Clear();
  lcd.BacklightOn(true);
  char Line1[LCD_CHAR_COLUMNS + 1];
  char Line2[LCD_CHAR_COLUMNS + 1];
  CreateText(Line1, Line2);
  graph.RenderData(true, m_CurrentData, m_CurrentData.m_iThresholdPercent, m_CurrentData.m_iBaseReadingBy4*4, Line1, Line2);
}

void DATA::PrintData()
{
  char Line1[LCD_CHAR_COLUMNS + 1];
  char Line2[LCD_CHAR_COLUMNS + 1];
  CreateText(Line1, Line2);
  graph.RenderData(false, m_CurrentData, m_CurrentData.m_iThresholdPercent, m_CurrentData.m_iBaseReadingBy4*4, Line1, Line2);
}

void DATA::ViewTest()
{
  // diagnostic view of the IMU readings
  char Line[LCD_CHAR_COLUMNS + 1];
  lcd.Clear();
  do
  {
    int X = imu.ReadCentiG(PIN_IMU_X);
    int Y = imu.ReadCentiG(PIN_IMU_Y);
    int Z = imu.ReadCentiG(PIN_IMU_Z);
    // bars
    lcd.BigFont(0, 0, imu.AsBar(X, Line, 'X'));
    lcd.BigFont(1, 0, imu.AsBar(Y, Line, 'Y'));
    lcd.BigFont(2, 0, imu.AsBar(Z, Line, 'Z'));
    
    // raw numbers
    Line[0] = 0;
    AppendCentiG(Line, X);
    AppendCentiG(Line, Y);
    AppendCentiG(Line, Z);
    lcd.BigFont(3, 0, Line);
  } while (!btn1Green.CheckButtonPress() && !btn2Red.CheckButtonPress());
}

bool DATA::TakeReading()
{
  // take a reading of "activity" from the IMU
  // false means stop
  int X = imu.ReadRaw(PIN_IMU_X);
  int Y = imu.ReadRaw(PIN_IMU_Y);
  int Z = imu.ReadRaw(PIN_IMU_Z);
  int dX = m_AvgX.DataPoint(X);
  int dY = m_AvgY.DataPoint(Y);
  int dZ = m_AvgZ.DataPoint(Z);
  // largest deviation from rolling averages:
  int dMax = max(dX, max(dY, dZ));
  m_iMaxInactivity = max(m_iMaxInactivity, dMax);

  // deviation as a percentage of part of the theoretical range (0..1023)
  int Percent = dMax*100/(m_CurrentData.m_iBaseReadingBy4*4);
  if (Percent < m_CurrentData.m_iThresholdPercent) // below threshold for movement
  {
    m_iInactivityWindowSum += Percent;
    m_iInactivityWindowCount++;
    Percent = 0;
  }
  if (Percent > 100)
    Percent = 100;

  if (Percent > 0)
  {
    m_iSamplePercentSum += Percent;
    m_iSamplePercentCount++;
  }
  
  byte NowMinutes = rtc.ReadMinute();
  m_iLastReadingMS = millis();
  if (m_iLastReadingMinute != NowMinutes)
  {
    // every minute
    m_iReadingMinutes++;
    m_iLastReadingMinute = NowMinutes;
    if (m_iReadingMinutes >= DATA_SAMPLE_MINUTES)
    {
      // every 6 minutes
      m_iReadingMinutes = 0;
      m_CurrentData.m_pActivity[m_CurrentData.m_iSampleCount++] = CalcDataValue();
    }
  }
  return (m_CurrentData.m_iSampleCount < DATA_MAX_SAMPLES) && !CheckAutoEnd(NowMinutes);
}

byte DATA::CalcDataValue()
{
  // compute the data
  long SamplePercent = 0;
  if (m_iSamplePercentCount)
  {
    long Count = m_iSamplePercentCount;
    SamplePercent = m_iSamplePercentSum/Count;
  }
  else
  {
    if (false)
    {
      // max movement below threshold
      SamplePercent = min(m_CurrentData.m_iThresholdPercent - 1, m_iMaxInactivity*100/(m_CurrentData.m_iBaseReadingBy4*4));
    }
    else
    {
      // avg non movement?
      if (m_iInactivityWindowCount)
      {
        double dbl = (double)m_iInactivityWindowSum/(double)m_iInactivityWindowCount + 0.5;
        SamplePercent = min(m_CurrentData.m_iThresholdPercent - 1, (int)dbl);
      }
    }
  }
    
  // reset
  m_iSamplePercentSum = 0;
  m_iSamplePercentCount = 0;

  m_iMaxInactivity = 0;
  m_iInactivityWindowSum = 0;
  m_iInactivityWindowCount = 0;
  
  return SamplePercent;
}

bool DATA::CheckAutoEnd(byte Minute)
{
  // check if logging should auto-end now
  if (cfg.m_AutoOn && (Minute % 15) == 0)
  {
    bool bWeekDay = (rtc.ReadDayOfWeek() > 2);
    // only check on the quarters
    byte Hour;
    byte Quarter;
    cfg.UnPackTime(bWeekDay?cfg.m_WeekOff:cfg.m_WeekendOff, Hour, Quarter);
    rtc.ReadTime();
    // and only an exact match
    return (Hour == rtc.m_Hour24 && Quarter*15 == rtc.m_Minute);
  }
  return false;
}

void DATA::ResetEEPROM()
{
  // erase all pages in flash
  for (int Page = 0; Page < DATA_NUM_PAGES; Page++)
  {
    int Base = Page*sizeof(tDayData);
    for (int Idx = 0; Idx < sizeof(tDayData); Idx++)
    {
      EEPROM.write(Base + Idx, 0xFF);
    }
  }
}

void DATA::WriteData()
{
  // write the current page to flash
  int Base = m_iPageIndex*sizeof(tDayData);
  byte* pData = (byte*)&m_CurrentData;
  for (int Idx = 0; Idx < sizeof(tDayData); Idx++, pData++)
  {
    EEPROM.write(Base + Idx, *pData);
  }
}

bool DATA::ReadData(int Day)
{
  // read the page from flash
  int Base = Day*sizeof(tDayData);
  byte* pData = (byte*)&m_CurrentData;
  for (int Idx = 0; Idx < sizeof(tDayData); Idx++, pData++)
  {
    *pData = EEPROM.read(Base + Idx);
  }
  if (IsValid(m_CurrentData.m_iBaseReadingBy4, m_CurrentData.m_iThresholdPercent))
  {
    m_iPageIndex = Day;
    return true;
  }
  else
  {
    m_iPageIndex = -1;
    return false;
  }
}

int DATA::GetLatestDayPage()
{
  // search the pages in flash for the one with the latest date
  int LatestDayPage = -1;  // none
  byte LatestDate = 0;
  byte LatestMonth = 0;
  for (int Page = 0; Page < DATA_NUM_PAGES; Page++)
  {
    int Idx = Page*sizeof(tDayData);
    byte Base = EEPROM.read(Idx++);
    byte Threshold = EEPROM.read(Idx++);
    if (IsValid(Base, Threshold))
    {
      byte ThisDate = EEPROM.read(Idx++);
      byte ThisMonth = EEPROM.read(Idx++);
      if (ThisMonth > LatestMonth ||
          (ThisMonth == LatestMonth && ThisDate > LatestDate))
      {
        LatestMonth = ThisMonth;
        LatestDate = ThisDate;
        LatestDayPage = Page;
      }
    }
  }
  return LatestDayPage;
}

void DATA::CreateText(char* Line1, char* Line2)
{
  // format the text lines which annotate the graph
  
  // date and start/end time
  // 1234567890123456
  // Mon 13 hhmm-hhmm
  char Temp[22];
  strcpy(Temp, "SatSunMonTueWedThuFri");
  strcpy(Line1, Temp + (m_CurrentData.m_iStartDayOfWeek - 1)*3);
  Line1[3] = 32;
  Line1[4] = 0;
  GRAPH::AppendNum(m_CurrentData.m_iStartDate, Line1, ' ');
  strcat(Line1, " ");
  GRAPH::AppendNum(m_CurrentData.m_iStartHour24, Line1, '0');
  GRAPH::AppendNum(m_CurrentData.m_iStartMinute, Line1, '0');
  int ElapsedMinutes = m_CurrentData.m_iSampleCount*DATA_SAMPLE_MINUTES;
  int ElapsedHours = ElapsedMinutes / 60;
  ElapsedMinutes = ElapsedMinutes % 60;
  int EndHour = m_CurrentData.m_iStartHour24 + ElapsedHours;
  int EndMinute = m_CurrentData.m_iStartMinute + ElapsedMinutes;
  EndHour += (EndMinute / 60);
  EndMinute %= 60;
  EndHour %= 24;
  strcat(Line1, "-");
  GRAPH::AppendNum(EndHour, Line1, '0');
  GRAPH::AppendNum(EndMinute, Line1, '0');
  
  // "sleep" stats
  // 1234567890123456
  // T=HH.Hh M=HH.Hh  
  int TotalHours;
  int TotalMin;
  int MaxHour;
  int MaxMin;
  GetStats(TotalHours, TotalMin, MaxHour, MaxMin);
  strcpy(Line2, "T=");
  GRAPH::AppendNum(TotalHours, Line2, ' ');
  strcat(Line2, ".");
  GRAPH::AppendNum((10*TotalMin)/60, Line2, '\x0');
  strcat(Line2, "h M=");
  GRAPH::AppendNum(MaxHour, Line2, ' ');
  strcat(Line2, ".");
  GRAPH::AppendNum((10*MaxMin)/60, Line2, '\x0');
  strcat(Line2, "h");
}

void DATA::GetStats(int& TotalHours, int& TotalMin, int& MaxHour, int& MaxMin)
{
  // total amount of inactivity ("sleep")
  // maximum contiguous period of inactivity
  int TotalCount = 0;
  int MaxCount = 0;
  int Max = 0;
  for (int Idx = 0; Idx < m_CurrentData.m_iSampleCount; Idx++)
  {
    if (m_CurrentData.m_pActivity[Idx] < m_CurrentData.m_iThresholdPercent)
    {
      TotalCount++;
      Max++;
    }
    else
    {
      if (Max > MaxCount)
        MaxCount = Max;
      Max = 0;
    }
  }
  TotalHours = (TotalCount*DATA_SAMPLE_MINUTES) / 60;
  TotalMin = (TotalCount*DATA_SAMPLE_MINUTES) % 60;
  MaxHour = (MaxCount*DATA_SAMPLE_MINUTES) / 60;
  MaxMin = (MaxCount*DATA_SAMPLE_MINUTES) % 60;
}

bool DATA::PageIsToday(int Page)
{
  // true if the given page has the same date as the current one
  int Idx = Page*sizeof(tDayData);
  byte Base = EEPROM.read(Idx++);
  byte Threshold = EEPROM.read(Idx++);
  if (IsValid(Base, Threshold))
  {
    byte ThisDate = EEPROM.read(Idx++);
    byte ThisMonth = EEPROM.read(Idx++);
    return (ThisDate == m_CurrentData.m_iStartDate && ThisMonth == m_CurrentData.m_iStartMonth);
  }
  return false;
}

void DATA::AppendCentiG(char* Buffer, int CentiG)
{
  // append CentiG as a string to buffer, "+/-N.N "
  if (CentiG < 0)
    strcat(Buffer, "-");
  else
    strcat(Buffer, "+");
  CentiG = abs(CentiG)/10;
  GRAPH::AppendNum(CentiG/10, Buffer, '\x0');  
  strcat(Buffer, ".");
  GRAPH::AppendNum(CentiG%10, Buffer, '\x0');  
  strcat(Buffer, " ");
}

bool DATA::IsValid(byte& Base, byte& Threshold)
{
  return (BASE_MIN <= Base && Base <= BASE_MAX && THRESH_MIN <= Threshold && Threshold <= THRESH_MAX);
}

void DATA::DumpCurrentPage()
{
  Serial.print(" Base   ");Serial.println((int)m_CurrentData.m_iBaseReadingBy4);
  Serial.print(" Thresh ");Serial.println((int)m_CurrentData.m_iThresholdPercent);
  Serial.print(" Date   ");Serial.println((int)m_CurrentData.m_iStartDate);
  Serial.print(" Month  ");Serial.println((int)m_CurrentData.m_iStartMonth);
  Serial.print(" Day    ");Serial.println((int)m_CurrentData.m_iStartDayOfWeek);
  Serial.print(" Hour   ");Serial.println((int)m_CurrentData.m_iStartHour24);
  Serial.print(" Min    ");Serial.println((int)m_CurrentData.m_iStartMinute);
  Serial.print(" Count  ");Serial.println((int)m_CurrentData.m_iSampleCount);
}

void DATA::DumpPages()
{
  for (int Day = 0; Day < DATA_NUM_PAGES; Day++)
  {
    int Base = Day*sizeof(tDayData);
    byte* pData = (byte*)&m_CurrentData;
    for (int Idx = 0; Idx < sizeof(tDayData); Idx++, pData++)
    {
      *pData = EEPROM.read(Base + Idx);
    }
    Serial.print("Page ");Serial.println(Day);
    DumpCurrentPage();
  }
}

