#include <Arduino.h>
#include "GRAPH.h"
#include "LCD.h"
#include "FONT.h"
#include "Printer.h"

GRAPH graph;

#define kMargin 1

void GRAPH::Init()
{
  lcd.ShowGraphics(false);
  lcd.ClearGraphics();
  m_bSolidFill = true;
  m_iBandHeight = 4;
}

char* GRAPH::AppendNum(int Num, char* Buffer, char Lead)
{
  // just 0..999
  int Len = strlen(Buffer);
  if (Num >= 100)
  {
    if (Lead)
      Lead = '0';
    Buffer[Len++] = '0' + (Num / 100);
  }
  Num = Num % 100;
  if (Num < 10)
  {
    if (Lead)
      Buffer[Len++] = Lead;
  }
  else
  {
    Buffer[Len++] = '0' + (Num / 10);
  }
  Buffer[Len++] = '0' + (Num % 10);
  Buffer[Len] = 0;
  return Buffer;
}

void GRAPH::RenderData(bool ToScreen, DATA::tDayData& Data, byte ThresholdPercent, byte Reference, char* Line1, char* Line2)
{
  tByteBand::tBandLayout Layout;
  if (ToScreen)
  {
    m_iBandHeight = 4;
    Layout = tByteBand::eHorizontalBytes;
    lcd.BigFont(0, 0, Line1);
    lcd.BigFont(1, 0, Line2);
    lcd.ShowGraphics(true);
  }
  else
  {
    m_iBandHeight = 5;
    Layout = tByteBand::eVerticalBytes;
    printer.StartPrint();
    printer.Println(Line1, true);
    printer.Println(Line2, true);
  }
  
  for (int Band = m_iBandHeight - 1; Band >= 0; Band--)
  {
    tByteBand::tPixelPos Pos;
    int DataBand;
    byte DataRow;
    band.Init(Band, Layout);
    
    // Y-axis
    if (band.GetPixel(kByteBandHeight - 1, kMargin - 1, Pos))
    {
      do
        band.SetPixel(Pos, true);
      while (band.NextPixelDown(Pos));
    }
    
    int MinCounter =  Data.m_iStartMinute;
    int HourCounter = Data.m_iStartHour24;
    // activity
    for (int col = 0; col < Data.m_iSampleCount; col++) 
    {
      if (GetPos(band, Data.m_pActivity[col], kMargin + col, Pos, DataBand, DataRow))
      {
        band.SetPixel(Pos, true);
        while (band.NextPixelDown(Pos))
          band.SetPixel(Pos, FillPattern(Pos));
      }
      else if (DataBand > Band && band.GetPixel(kByteBandHeight - 1, kMargin + col, Pos))
      {
        // fill the rest of the band
        do
          band.SetPixel(Pos, FillPattern(Pos));
        while (band.NextPixelDown(Pos));
      }
      
      if (MinCounter < DATA_SAMPLE_MINUTES)
      {
        bool LabelHour = (HourCounter % 3 == 0) && (Band == m_iBandHeight - 1);
        // vt lines
        if ((Band == m_iBandHeight - 1) && band.GetPixel(LabelHour?(kByteBandHeight - FONT_HEIGHT - 2):(kByteBandHeight - 1), kMargin + col, Pos))
        {
          do
            SetLinePixel(band, Pos);
          while (band.NextPixelDown(Pos));
        }
        if (LabelHour)
        {
          char HourStr[4];
          HourStr[0] = 0;
          int HourLabel = HourCounter;
          //if (HourLabel > 12)
          //  HourLabel -= 12;
          AppendNum(HourLabel, HourStr, '\0');
          FONT::SmallFont(band, kByteBandHeight - FONT_HEIGHT - 1, kMargin + col - strlen(HourStr)*(FONT_WIDTH + 1)/2, HourStr);
        }
      }
      MinCounter += DATA_SAMPLE_MINUTES;
      if (MinCounter >= 60)
      {
        MinCounter = 0;
        HourCounter++;
        HourCounter %= 24;
      }
    }

    char Label[5];
    
    // threshold (horizontal)
    if (GetPos(band, ThresholdPercent, kMargin, Pos, DataBand, DataRow))
    {
      do
        SetLinePixel(band, Pos);
      while (band.NextPixelRight(Pos) && Pos.m_iCol < (Data.m_iSampleCount + kMargin));
      Label[0] = 0;
      AppendNum(ThresholdPercent, Label, '\0');
      strcat(Label, "%");
      FONT::SmallFont(band, 0, min(kByteBandWidth-strlen(Label)*(FONT_WIDTH + 1), Data.m_iSampleCount + kMargin), Label);
    }

    if (Band == m_iBandHeight - 1)
    {
      strcpy(Label, "@");
      AppendNum(Reference, Label, '\0');
      FONT::SmallFont(band, 0, min(kByteBandWidth-strlen(Label)*(FONT_WIDTH + 1), Data.m_iSampleCount + kMargin), Label);
    }
    
    // send the band
    if (ToScreen)
    {
      lcd.ShowBand((kByteBandHeight - 1 - Band)*kByteBandHeight, band.Data());
    }
    else
    {
      printer.PrintBandDouble(band.Data(), kByteBandWidth);
    }
  }
  
  if (ToScreen)
  {
    lcd.ShowGraphics(true);
  }
  else
  {
    printer.FeedLines(2);
  }
}

bool GRAPH::GetPos(tByteBand& band, byte Data, byte Col, tByteBand::tPixelPos& Pos, int& DataBand, byte& DataRow)
{
  int ScaledData = (m_iBandHeight*kByteBandHeight*Data) / 100;  
  DataBand = ScaledData / kByteBandHeight;
  DataRow = ScaledData % kByteBandHeight;
  return (DataBand == band.Band() && band.GetPixel(DataRow, Col, Pos));
}

bool GRAPH::FillPattern(const tByteBand::tPixelPos& Pos)
{
  return ((Pos.m_iRow + Pos.m_iCol) % 2) == 0 || m_bSolidFill;
}

void GRAPH::SetLinePixel(tByteBand& band, const tByteBand::tPixelPos& Pos)
{
  if (m_bSolidFill)
  {
    band.InvertPixel(Pos);
  }
  else
  {
    band.SetPixel(Pos, true);
  }
}
