#include <Arduino.h>
#include "LCD.h"

LCD12864 lcd;

LCD12864::LCD12864():
  m_iLatchPin(0xFF),
  m_iClockPin(0xFF),
  m_iDataPin(0xFF),
  m_bBacklightOn(true)
{
}

void LCD12864::Init(byte LatchPin, byte ClockPin, byte DataPin)
{
  // set up the pins to talk to the 595
  m_iLatchPin = LatchPin;
  m_iClockPin = ClockPin;
  m_iDataPin = DataPin;
  m_bBacklightOn = true;
  pinMode(m_iLatchPin, OUTPUT);
  pinMode(m_iClockPin, OUTPUT);
  pinMode(m_iDataPin, OUTPUT);
 
  // select 4-bit, do it twice becase the first time the 4 lower bits will be random
  Command(LCD_FUNCTION + LCD_FUNCTION_4BIT + LCD_FUNCTION_BASIC);
  Command(LCD_FUNCTION + LCD_FUNCTION_4BIT + LCD_FUNCTION_BASIC);
  Command(LCD_CLEAR);
  Command(LCD_DISPLAY + LCD_DISPLAY_ON + LCD_DISPLAY_CURSOR_OFF+ LCD_DISPLAY_BLINK_OFF);
  Command(LCD_ENTRY + LCD_ENTRY_CURSOR_RIGHT);
}

void LCD12864::Clear()
{
  Command(LCD_FUNCTION + LCD_FUNCTION_4BIT + LCD_FUNCTION_BASIC);
  Command(LCD_CLEAR);
}

void LCD12864::BacklightOn(bool On)
{
  if (m_bBacklightOn != On)
  {
    m_bBacklightOn = On;
    // force an update but without sending a commant to the LCD
    ShiftOut(0);
  }
}

bool LCD12864::BacklightIsOn()
{
  return m_bBacklightOn;
}


void LCD12864::BigFont(byte Row, byte Col, char* Str, int Len /*= -1*/)
{
  byte Addr = 0;
  Row &= B00000011;
  Col &= B00000111;
  Addr = (Row % 2)?0x90:0x80;
  if (Row > 1) 
    Addr += 0x08;
  Addr += Col;
  ExtendedMode(false);
  Command(Addr);
  if (Len == -1)
    Len = strlen(Str);
  while (Len)
  {
    Data(*Str);
    Str++;
    Len--;
  }
}

void LCD12864::InvertLine(byte Row)  // DOES NOT WORK, 0 inverts 0 & 2...
{
  ExtendedMode(true);
  Command(LCD_INVERT | (Row & B11));
}

void LCD12864::ExtendedMode(bool Extended)
{
  if (Extended)
  {
    Command(LCD_FUNCTION + LCD_FUNCTION_4BIT + LCD_FUNCTION_EXTENDED);
  }
  else
  {
    Command(LCD_FUNCTION + LCD_FUNCTION_4BIT + LCD_FUNCTION_BASIC);
  }
}

void LCD12864::ShowGraphics(bool Show)
{
  ExtendedMode(true);
  if (Show)
  {
    Command(LCD_EXT_FUNCTION + LCD_EXT_FUNC_4BIT + LCD_EXT_FUNC_EXTENDED + LCD_EXT_FUNC_GRAPH_ON);
  }
  else
  {
    Command(LCD_EXT_FUNCTION + LCD_EXT_FUNC_4BIT + LCD_EXT_FUNC_EXTENDED + LCD_EXT_FUNC_GRAPH_OFF);
  }
}

void LCD12864::GetGraphicsAddress(byte Row, byte Col, byte& AddrA, byte& AddrB)
{
  if (Row < 32)
  {
    AddrA = LCD_GDRAM_ADDR + Row;
    AddrB = LCD_GDRAM_ADDR + Col;
  }
  else
  {
    AddrA = LCD_GDRAM_ADDR + Row - 32;
    AddrB = LCD_GDRAM_ADDR + 8 + Col;
  }
}
void LCD12864::SetGraphicsAddress(byte Row, byte Col)
{
  ExtendedMode(true);
  byte AddrA;
  byte AddrB;
  GetGraphicsAddress(Row, Col, AddrA, AddrB);
  Command(AddrA);
  Command(AddrB);
}

void LCD12864::ShowBand(byte TopRow, byte* pData)
{
  for (int RowOffset = 0; RowOffset < 8; RowOffset++)
  {
    SetGraphicsAddress(TopRow + RowOffset, 0);
    for (int ColOffset = 0; ColOffset < 16; ColOffset++)
    {
      Data(*pData);
      pData++;
    }
  }
}

void LCD12864::ClearGraphics(byte StartRow /*= 0*/, byte EndRow /*= 63*/)
{
  ExtendedMode(true);
  for (int Row = StartRow; Row <= EndRow; Row++)
  {
    byte AddrA;
    byte AddrB;
    GetGraphicsAddress(Row, 0, AddrA, AddrB);
    Command(AddrA);
    Command(AddrB);
    for (int Idx = 0; Idx < 16; Idx++)
      Data(0x00);
  }  
}

void LCD12864::Command(byte Cmd)
{
  SendByte(0, Cmd);
}

void LCD12864::Data(byte Dat)
{
  SendByte(1, Dat);
}


void LCD12864::ShiftOut(byte Byte)
{
  if (m_bBacklightOn)  // honour the backlight bit
    Byte |= bit(eBacklight_Bit);

  // take the latchPin low so the pins don't change while we are sending in bits:
  digitalWrite(m_iLatchPin, LOW);
  // shift out the bits:
  shiftOut(m_iDataPin, m_iClockPin, MSBFIRST, Byte);  
  //take the latch pin high so the pins change
  digitalWrite(m_iLatchPin, HIGH);
}

void LCD12864::StartCustomChar(byte Ch)
{
  lcd.ExtendedMode(true);
  lcd.Command(LCD_EXT_SCROOL_OR_RAM + 0);  // Scroll or RAM Address Select
  lcd.ExtendedMode(false);
  lcd.Command(LCD_CGRAM_ADDR + Ch*16);
}

void LCD12864::SendByte(byte RS, byte Byte) 
{
  byte valueToShift = 0;

  bitWrite(valueToShift, eEN_Bit, 1);
  ShiftOut(valueToShift);  
  delayMicroseconds(125);

  bitWrite(valueToShift, eRS_Bit, RS);
  // upper 4 bits
  bitWrite(valueToShift, eD4_Bit, bitRead(Byte, 4));
  bitWrite(valueToShift, eD5_Bit, bitRead(Byte, 5));
  bitWrite(valueToShift, eD6_Bit, bitRead(Byte, 6));
  bitWrite(valueToShift, eD7_Bit, bitRead(Byte, 7));
  ShiftOut(valueToShift);  
  delayMicroseconds(50);

  bitWrite(valueToShift, eEN_Bit, 0);//  digitalWrite(EN, 0);   
  ShiftOut(valueToShift);  
  delayMicroseconds(50);

  bitWrite(valueToShift, eEN_Bit, 1);//  digitalWrite(EN,1);  
  valueToShift &= B11100001;
  // lower 4 bits
  bitWrite(valueToShift, eD4_Bit, bitRead(Byte, 0));
  bitWrite(valueToShift, eD5_Bit, bitRead(Byte, 1));
  bitWrite(valueToShift, eD6_Bit, bitRead(Byte, 2));
  bitWrite(valueToShift, eD7_Bit, bitRead(Byte, 3));
  ShiftOut(valueToShift);  
  delayMicroseconds(50);  // Data set-up time tDSW?

  bitWrite(valueToShift, eEN_Bit, 0);//  digitalWrite(EN, 0);   
  ShiftOut(valueToShift);  
  delayMicroseconds(100);
  if (RS == 0 && Byte == LCD_CLEAR)  // Command just clear??
    delayMicroseconds(250);  // worst case?? after init?
}

