#include <Arduino.h>
#include <avr/pgmspace.h>
#include "MENU.h"
#include "PINS.h"
#include "LCD.h"
#include "Button.h"
#include "RTC.h"
#include "CFG.h"
#include "GRAPH.h"
#include "Printer.h"

Menu menu;

// turn the backlight off after this much inactivity
#define IDLE_TIMEOUT_MS (unsigned long)(5L*60L*1000L)  // 5 min

Menu::Menu()
{
}

void Menu::Init()
{
  m_iCurrentlyInvertedLine = -1;
  m_iButtonTimeMS = millis();
  
  lcd.StartCustomChar(0);
  for (int Idx = 0; Idx < 32; Idx++)
  {
    lcd.Data(pgm_read_byte(m_pCustomPointerChar + Idx));
  }
}

Menu::tBtn Menu::WaitForButtonOrTimeout(long TimeoutMS /*= 0*/)
{
  // waits for a btn or timeout (eNone) if non-zero
  unsigned long StartMS = millis();
  do
  {
    if (btn1Green.CheckButtonPress()) return eGreen;
    if (btn2Red.CheckButtonPress()) return eRed;
    if (TimeoutMS && (millis() - StartMS) >= (unsigned long)abs(TimeoutMS))
    {
      return eNone;
    }
  }
  while (true);
}

Menu::tBtn Menu::WaitForButton(unsigned long StartMS, bool CheckTime)
{
  tBtn Btn = eNone;
  do
  {
    if (btn1Green.CheckButtonPress()) 
    {
      if (btn2Red.IsDown())
      {
        Btn = eGreenWhileRedDown;
      }
      else
      {
        Btn = eGreen;
      }
    }
    else if (btn2Red.CheckButtonPress())
    {
      Btn = eRed;
    }
    unsigned long Now = millis();
    if (CheckTime)
    {
      // bail periodically
      if ((Now - StartMS) >= 15000L)
      {
        return eTimeCheck;
      }
    }
    
    if (Btn == eNone)
    {
      if (((Now - m_iButtonTimeMS) >= IDLE_TIMEOUT_MS))
      {
        // no activity, b/l off
        lcd.BacklightOn(false);
        // AND clear the screen
        lcd.Clear();
      }
    }
    else
    {
      if (!lcd.BacklightIsOn())
      {
        // btn just wakes up b/l
        m_iButtonTimeMS = Now;
        lcd.BacklightOn(true);
        // force a redraw
        return eRedraw;
      }
    }
  } while (Btn == eNone);
  lcd.BacklightOn(true);
  m_iButtonTimeMS = millis();
  return Btn;
}

void Menu::IncrementItem(byte& Item, int Min, int Max, bool Increment, int Step)
{
  if (Increment)
  {
    Item += Step;
  }
  else
  {
    if (Item < Min)
      Item = Min;
  }
  if (Item > Max)
    Item = Min;
}

char* Menu::StdItem(byte& Value, int Min, int Max, bool IncrementFirst, char* Buffer, char Lead, int Step /*= 1*/)
{
  IncrementItem(Value, Min, Max, IncrementFirst, Step);
  return GRAPH::AppendNum(Value, Buffer, Lead);
}

char* Menu::AutoItem(byte& Value, bool IncrementFirst, char* Buffer, byte Min, byte Max)
{
  byte Hr;
  byte Quarter;
  if (Value < Min || Value > Max)
  {
    Value = Min;
  }
  cfg.UnPackTime(Value, Hr, Quarter);
  if (IncrementFirst)
  {
    if (Value == Max)
    {
      Value = Min;
      cfg.UnPackTime(Value, Hr, Quarter);
    }
    else
    {
      Quarter++;
      if (Quarter > 3)
      {
        Quarter = 0;
        Hr++;
      }
      if (Hr > 23)
      {
        Hr = 0;
      }
      Value = cfg.PackTime(Hr, Quarter);
    }
  }
  GRAPH::AppendNum(Hr, Buffer, '0');
  return GRAPH::AppendNum(Quarter*15, Buffer, '0');
}

char* Menu::AutoOnItem(byte& Value, bool IncrementFirst, char* Buffer)
{
  if (IncrementFirst)
  {
    if (Value) 
      Value = 0;
    else
      Value = 1;
  }
  else
  {
    // normalize
    if (Value) 
      Value = 1;
  }
  if (Value) 
    strcpy(Buffer, "On");
  else
    strcpy(Buffer, "Off");
  return Buffer;
}

bool Menu::GetItem(tMenuItems Item, char* LineBuffer, bool IncrementFirst)
{
  // fills-in LineBuffer incrementing the setting first if required.  returns true if the item is a menu item
  bool IsMenuItem = false;
  char* pValue = NULL;
  char ValueBuffer[10];
  LoadString(Item, LineBuffer);
  strcpy(ValueBuffer, "");
  switch (Item)
  {
    case eStartMenu:
    case eViewMenu:
    case ePrintMenu:
    case eConfigMenu:
    case eViewLastNightMenu:
    case eViewAllMenu:
    case eViewTestMenu:
    case eBackMenu:
    case eSetMenu:
    case ePrintLastNightMenu:
    case ePrintAllMenu:
    case ePrintTestMenu:
    case eConfigTimeMenu:
    case eConfigDateMenu:
    case eConfigAutoMenu:
    case eConfigSettingsMenu:
    {
      IsMenuItem = true;
      break;
    }

    case eCurrentHourItem:
    {
      pValue = StdItem(rtc.m_Hour24, 0, 23, IncrementFirst, ValueBuffer, '0');
      break;
    }
    case eCurrentMinuteItem:
    {
      pValue = StdItem(rtc.m_Minute, 0, 59, IncrementFirst, ValueBuffer, '0');
      break;
    }
    case eCurrentDayItem:
    {
      pValue = StdItem(rtc.m_DayOfWeek, 1, 7, IncrementFirst, ValueBuffer, '\0');
      break;
    }

    case eCurrentDateItem:
    {
      pValue = StdItem(rtc.m_DayOfMonth, 1, 31, IncrementFirst, ValueBuffer, ' ');
      break;
    }
    case eCurrentMonthItem:
    {
      pValue = StdItem(rtc.m_Month, 1, 12, IncrementFirst, ValueBuffer, ' ');
      break;
    }
    case eCurrentYearItem:
    {
      pValue = StdItem(rtc.m_Year, 12, 20, IncrementFirst, ValueBuffer, ' ');
      break;
    }
    
    case eAutoWeekdayMenu:
    case eAutoWeekendMenu:
    {
      IsMenuItem = true;
      break;
    }

    case eAutoOnItem:
    {
      pValue = AutoOnItem(cfg.m_AutoOn, IncrementFirst, ValueBuffer);
      break;
    }
    case eAutoWeekdayOnItem:
    {
      pValue = AutoItem(cfg.m_WeekOn, IncrementFirst, ValueBuffer, cfg.PackTime(21, 0), cfg.PackTime(23, 3));  // 9:00-11:45pm
      break;
    }
    case eAutoWeekdayOffItem:
    {
      pValue = AutoItem(cfg.m_WeekOff, IncrementFirst, ValueBuffer, cfg.PackTime(6, 0), cfg.PackTime(9, 3));  // 6:00-9:45am
      break;
    }
    case eAutoWeekendOnItem:
    {
      pValue = AutoItem(cfg.m_WeekendOn, IncrementFirst, ValueBuffer, cfg.PackTime(21, 0), cfg.PackTime(23, 3));  // 9:00-11:45pm
      break;
    }
    case eAutoWeekendOffItem:
    {
      pValue = AutoItem(cfg.m_WeekendOff, IncrementFirst, ValueBuffer, cfg.PackTime(9, 0), cfg.PackTime(11, 3));  // 9:00-11:45am
      break;
    }
    case eSettingBaseReading:
    {
      IncrementItem(cfg.m_iBaseReadingBy4, BASE_MIN, BASE_MAX, IncrementFirst, 4);
      pValue = GRAPH::AppendNum(cfg.m_iBaseReadingBy4*4, ValueBuffer, ' ');
      strcat(pValue, "/1024");
      break;
    }
    case eSettingThresholdPercent:
    {
      pValue = StdItem(cfg.m_iThresholdPercent, THRESH_MIN, THRESH_MAX, IncrementFirst, ValueBuffer, ' ', 5);
      strcat(pValue, "%");
      break;
    }
  }

  if (pValue)
  {
    strcat(LineBuffer, ":");
    strcat(LineBuffer, pValue);
  }

  // pad to full width
  while (strlen(LineBuffer) < LCD_CHAR_COLUMNS - 2)  
  {
    strcat(LineBuffer, " ");
  }
  return IsMenuItem;
}

void Menu::SelectItem(int Line)
{
  if (Line != -1)
  {
    if (m_iCurrentlyInvertedLine != -1 && Line != m_iCurrentlyInvertedLine)
    {
      lcd.BigFont(m_iCurrentlyInvertedLine, 0, "  ");
    }
    int CustomChar0000 = 0;
    lcd.BigFont(Line, 0, (char*)&CustomChar0000, 2);//=>");
    m_iCurrentlyInvertedLine = Line;
  }
  else
  {
    if (m_iCurrentlyInvertedLine != -1)
    {
      lcd.BigFont(m_iCurrentlyInvertedLine, 0, "  ");
    }
    m_iCurrentlyInvertedLine = -1;
  }
}

Menu::tMenuItems Menu::DoMenu(tMenuItems M0, tMenuItems M1, tMenuItems M2, tMenuItems M3, int& SelectedLine)
{
  unsigned long StartMS = millis();
  tMenuItems pItems[LCD_CHAR_ROWS];
  char LineBuffer[LCD_CHAR_COLUMNS + 1];
  pItems[0] = M0;
  pItems[1] = M1;
  pItems[2] = M2;
  pItems[3] = M3;
  int Col = 1;
  strcpy(LineBuffer, "");
  for (int Line = 0; Line < LCD_CHAR_ROWS; Line++)
  {
    GetItem(pItems[Line], LineBuffer, false);
    lcd.BigFont(Line, Col, LineBuffer);
  }
  SelectItem(SelectedLine);
  do 
  {
    tBtn Btn = WaitForButton(StartMS, M0 == eStartMenu);
    if (Btn == eRed)
    {
      SelectedLine++;
      if (SelectedLine >= LCD_CHAR_ROWS || pItems[SelectedLine] == eNULL_ITEM)
      {
        SelectedLine = 0;
      }
      SelectItem(SelectedLine);
    }
    else if (Btn == eGreen)
    {
      if (GetItem(pItems[SelectedLine], LineBuffer, true))
      {
        // a menu item, get out
        return pItems[SelectedLine];
      }
      else
      {
        lcd.BigFont(SelectedLine, Col, LineBuffer);
      }
    }
    else if (Btn == eTimeCheck)
    {
      return eTIME_CHECK;
    }
    else if (Btn == eRedraw)
    {
      return eREDRAW;
    }
    else if (Btn == eGreenWhileRedDown && M0 == eStartMenu)
    {
      return eGREEN_RED;
    }
  } while (true);
}

void Menu::DoSubMenu(tMenuItems M0, tMenuItems M1, tMenuItems M2, tMenuItems M3)
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(M0, M1, M2, M3, SelectedLine);
  } while (Item != eSetMenu);
}


void Menu::DoViewMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(eViewLastNightMenu,
                  eViewAllMenu,
                  eViewTestMenu,
                  eBackMenu, SelectedLine);
    switch (Item)
    {
      case eViewLastNightMenu:
        if (data.LoadLast())
        {
          data.GraphData();
          WaitForButtonOrTimeout();
          lcd.ShowGraphics(false);
          lcd.Clear();
        }
        break;
      case eViewAllMenu:
        if (data.LoadLast())
        {
          do
          {
            data.GraphData();
            if (WaitForButtonOrTimeout() == Menu::eRed)
              break;
          } while (data.LoadPrev());
          lcd.ShowGraphics(false);
          lcd.Clear();
        }
        break;
      case eViewTestMenu:
        data.ViewTest();
        lcd.Clear();
        break;
    }
  } while (Item != eBackMenu);
}

void Menu::DoPrintMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(ePrintLastNightMenu,
                  ePrintAllMenu,
                  ePrintTestMenu,
                  eBackMenu, SelectedLine);
    switch (Item)
    {
      case ePrintLastNightMenu:
        if (data.LoadLast())
        {
          lcd.Clear();
          lcd.BigFont(0, 0, "Printing...");
          data.PrintData();
          printer.EndPrint();
          lcd.Clear();
        }
        break;
      case ePrintAllMenu:
        if (data.LoadLast())
        {
          lcd.BigFont(0, 0, "Printing...");
          do
          {
            data.PrintData();
          } while (data.LoadPrev());
          printer.EndPrint();
          lcd.Clear();
        }
        break;
      case ePrintTestMenu:
        printer.Println("Test Print", true);
        printer.FeedLines(8);
        break;
    }
  } while (Item != eBackMenu);
}

void Menu::DoConfigMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(eConfigTimeMenu,
                  eConfigDateMenu,
                  eConfigSettingsMenu,
                  eBackMenu, SelectedLine);
    switch (Item)
    {
      case eConfigTimeMenu:
        rtc.ReadTime();
        DoSubMenu(eCurrentHourItem,
                  eCurrentMinuteItem,
                  eCurrentDayItem, 
                  eSetMenu);
        rtc.WriteTime();
        break;
      case eConfigDateMenu:
        rtc.ReadTime();
        DoSubMenu(eCurrentDateItem,
                  eCurrentMonthItem,
                  eCurrentYearItem, 
                  eSetMenu);
        rtc.WriteTime();
        break;
      case eConfigSettingsMenu:
        DoSettingsMenu();
        break;
    }           
  } while (Item != eBackMenu);
}

void Menu::DoSettingsMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(eConfigAutoMenu,
                  eSettingBaseReading,
                  eSettingThresholdPercent,
                  eBackMenu, SelectedLine);
    switch (Item)
    {
      case eConfigAutoMenu:
        DoAutoMenu();
        break;
    }           
  } while (Item != eBackMenu);
  cfg.Save();
}

void Menu::DoAutoMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  do 
  {
    Item = DoMenu(eAutoOnItem,
                  eAutoWeekdayMenu,
                  eAutoWeekendMenu,
                  eBackMenu, SelectedLine);
    switch (Item)
    {
      case eAutoWeekdayMenu:
        DoSubMenu(eAutoWeekdayOnItem,
                  eAutoWeekdayOffItem,
                  eSetMenu,
                  eNULL_ITEM);
        cfg.Save();
        break;
      case eAutoWeekendMenu:
        DoSubMenu(eAutoWeekendOnItem,
                  eAutoWeekendOffItem,
                  eSetMenu, 
                  eNULL_ITEM);
        cfg.Save();
        break;
    }           
  } while (Item != eBackMenu);
}


Menu::tMenuItems Menu::DoMainMenu()
{
  int SelectedLine = 0;
  tMenuItems Item = eNULL_ITEM;
  lcd.ShowGraphics(false);
  lcd.Clear();
  m_iButtonTimeMS = millis();
  do 
  {
    Item = DoMenu(eStartMenu,
                  eViewMenu,
                  ePrintMenu,
                  eConfigMenu, SelectedLine);
    switch (Item)
    {
      case eViewMenu: 
        DoViewMenu(); 
        break;
      case ePrintMenu: 
        DoPrintMenu(); 
        break;
      case eConfigMenu: 
        DoConfigMenu(); 
        break;
      case eTIME_CHECK: 
        if (data.CheckAutoStart())
        {
          Item = eStartMenu;
        }
        break;
    }
  } while (Item != eStartMenu && Item != eGREEN_RED);
  return Item;
}

///////////////////////////////////////////////////////
// endless problems with PROGMEM...
// todo, F()?
// single string, delimiter is <lowercase><Uppercase>
//"Start""View""Print""Config""View Last""View All""View Test""Back""Set""Print Last""Print All""Config Time""Config Date""Config Auto""Hour""Minute""Day (1=Sat)""Date""Month""Year""Auto W'day""Auto W'end""Auto""W'day  On""W'day Off""W'end  On""W'end Off"
//StartViewPrintConfigView LastView AllView TestBackSetPrint LastPrint AllConfig TimeConfig DateConfig AutoHourMinuteDay (1=Sat)DateMonthYearAuto W'dayAuto W'endAutoW'day  OnW'day OffW'end  OnW'end Off

#define FU(s) (s)
void Menu::LoadString(tMenuItems Item, char* Buffer)
{
  switch (Item)
  {
    case eStartMenu:  strcpy(Buffer, FU("Start"));break;
    case eViewMenu:  strcpy(Buffer, FU("View\x10"));break;
    case ePrintMenu:  strcpy(Buffer, FU("Print\x10"));break;
    case eConfigMenu:  strcpy(Buffer, FU("Config\x10"));break;
    case eViewLastNightMenu:  strcpy(Buffer, FU("View Last"));break;
    case eViewAllMenu:  strcpy(Buffer, FU("View All"));break;
    case eViewTestMenu:  strcpy(Buffer, FU("View Test"));break;
    case eBackMenu:  strcpy(Buffer, FU("\x11""Back"));break;
    case eSetMenu:  strcpy(Buffer, FU("\x11Set"));break;
    case ePrintLastNightMenu:  strcpy(Buffer, FU("Print Last"));break;
    case ePrintAllMenu:  strcpy(Buffer, FU("Print All"));break;
    case ePrintTestMenu:  strcpy(Buffer, FU("Print Test"));break;
    case eConfigTimeMenu:  strcpy(Buffer, FU("Config Time\x10"));break;
    case eConfigDateMenu:  strcpy(Buffer, FU("Config Date\x10"));break;
    case eConfigAutoMenu:  strcpy(Buffer, FU("Config Auto\x10"));break;
    case eCurrentHourItem:  strcpy(Buffer, FU("Hour"));break;
    case eCurrentMinuteItem:  strcpy(Buffer, FU("Minute"));break;
    case eCurrentDayItem:  strcpy(Buffer, FU("Day (1=Sat)"));break;
    case eCurrentDateItem:  strcpy(Buffer, FU("Date"));break;
    case eCurrentMonthItem:  strcpy(Buffer, FU("Month"));break;
    case eCurrentYearItem:  strcpy(Buffer, FU("Year"));break;
    case eAutoWeekdayMenu:  strcpy(Buffer, FU("Auto W'day\x10"));break;
    case eAutoWeekendMenu:  strcpy(Buffer, FU("Auto W'end\x10"));break;
    case eAutoOnItem:  strcpy(Buffer, FU("Auto"));break;
    case eAutoWeekdayOnItem:  strcpy(Buffer, FU("W'day  On"));break;
    case eAutoWeekdayOffItem:  strcpy(Buffer, FU("W'day Off"));break;
    case eAutoWeekendOnItem:  strcpy(Buffer, FU("W'end  On"));break;
    case eAutoWeekendOffItem:  strcpy(Buffer, FU("W'end Off"));break;
    case eConfigSettingsMenu:  strcpy(Buffer, FU("Settings\x10"));break;
    case eSettingBaseReading:  strcpy(Buffer, FU("Base"));break;
    case eSettingThresholdPercent:  strcpy(Buffer, FU("Threshold"));break;
    
    default:  strcpy(Buffer, FU(""));break;
  }
}

// format is b7...b0 b7...b0
//           b7...b0 b7...b0
//           ...
prog_uchar Menu::m_pCustomPointerChar[] PROGMEM =
{
  B00000000, B10000000,
  B00000000, B11000000,
  B00000000, B11100000,
  B00000000, B11110000,
  B11111111, B11111000,
  B11111111, B11111100,
  B11111111, B11111110,
  B11111111, B11111111,
  B11111111, B11111110,
  B11111111, B11111100,
  B11111111, B11111000,
  B00000000, B11110000,
  B00000000, B11100000,
  B00000000, B11000000,
  B00000000, B10000000,
  B00000000, B00000000,

};

